# $Idaemons: /home/cvs/pkgtools/lib/pkgdb.rb,v 1.64 2002/01/15 20:04:17 knu Exp $

require 'singleton'
require 'tsort'
require "pkgmisc"

class PkgDB
  include Singleton
  include Enumerable

  DB_VERSION = [:FreeBSD, 4]
  # :db_version		=> DB_VERSION
  # :installed_pkgs	=> list of installed packages
  # ?pkgname		=> origin
  #   ...
  # /path/to/file	=> pkgname
  #   ...

  DB_DIR = ENV['PKG_DBDIR'] || '/var/db/pkg'

  PKGDB_FILES = {
    :contents => '+CONTENTS',
    :comment => '+COMMENT',
    :desc => '+DESC',
    :install => '+INSTALL',
    :post_install => '+POST-INSTALL',
    :deinstall => '+DEINSTALL',
    :post_deinstall => '+POST-DEINSTALL',
    :require => '+REQUIRE',
    :required_by => '+REQUIRED_BY',
    :display => '+DISPLAY',
    :mtree => '+MTREE_DIRS',
  }

  PREFIX = '/usr/local'

  CMD = {
    :pkg_add			=> '/usr/sbin/pkg_add',
    :pkg_delete			=> '/usr/sbin/pkg_delete',
    :pkg_info			=> '/usr/sbin/pkg_info',
    :make_describe_pass1	=> "#{PREFIX}/sbin/make_describe_pass1",
    :make_describe_pass2	=> "#{PREFIX}/sbin/make_describe_pass2",
    :pkg_fetch			=> "#{PREFIX}/sbin/pkg_fetch",
    :pkg_deinstall		=> "#{PREFIX}/sbin/pkg_deinstall",
    :pkg_fetch			=> "#{PREFIX}/sbin/pkg_fetch",
    :pkg_tarup			=> "#{PREFIX}/sbin/pkg_tarup",
    :pkgdb			=> "#{PREFIX}/sbin/pkgdb",
  }

  class DBError < StandardError
#    def message
#      "pkgdb error"
#    end
  end

  attr_reader :db_dir

  def initialize(*args)
    setup(*args)
  end

  def setup(db_dir = nil)
    @db_dir = File.expand_path(db_dir || DB_DIR)
    @abs_db_dir = nil

    @db_file = File.join(@db_dir, 'pkgdb.db')
    @db_filebase = @db_file.sub(/\.db$/, '')
    @db = nil
    @warned = false

    @installed_pkgs = nil

    self
  end

  def abs_db_dir
    unless @abs_db_dir
      Dir.chdir(@db_dir) {
	@abs_db_dir = Dir.pwd
      } rescue raise DBError, "Can't chdir to '#{@db_dir}'"
    end

    @abs_db_dir
  end

  def strip(path, installed_only = false)
    base = path.chomp('/')	# allow `pkgname/'

    if base.include?('/')
      dir, base = File.split(File.expand_path(base))

      if dir != @db_dir && dir != abs_db_dir
	return nil
      end
    end

    if installed_only && !installed?(base)
      return nil
    end

    base
  end

  def pkgdir(pkgname)
    File.join(@db_dir, pkgname)
  end

  def pkgdir?(dir)
    File.exist?(File.join(dir, PKGDB_FILES[:contents]))
  end

  def pkgfile(pkgname, filename)
    if filename.kind_of?(Symbol)
      filename = PKGDB_FILES[filename]
    end

    File.join(pkgdir(pkgname), filename)
  end

  def pkg(pkgname)
    installed?(pkgname) and PkgInfo.new(pkgname)
  rescue => e
    return nil
  end

  alias [] pkg

  def origin(pkgname)
    if unavailable?
      if p = pkg(pkgname)
	p.origin!
      else
	nil
      end
    else
      open_db

      @db['?' + pkgname]
    end
  rescue => e
    raise DBError, e.message
  end

  def set_origin(pkgname, origin)
    open_db

    @db['?' + pkgname] = origin
  rescue => e
    raise DBError, e.message
  end

  def deorigin(origin)
    if unavailable?
      list = glob().select { |pkg| pkg.origin == origin }

      if list.empty?
	nil
      else
	list
      end
    else
      open_db

      if str = @db['?' + origin]
	str.split
      else
	nil
      end
    end
  rescue => e
    raise DBError, e.message
  end

  def date_db_dir
    File.mtime(@db_dir) rescue Time.at(0)
  end

  def date_db_file
    if @db
      mtime = Marshal.load(@db[':mtime'])
    elsif File.exist?(@db_file)
      db = DBM.open(@db_filebase)
      mtime = Marshal.load(db[':mtime'])
      db.close
    else
      mtime = Time.at(0)
    end

    mtime
  rescue => e
    return Time.at(0)
  end

  def up_to_date?
    date_db_file() >= date_db_dir()
  end

  def check_db_version
    db_version = Marshal.load(@db[':db_version'])

    db_version[0] == DB_VERSION[0] && db_version[1] >= DB_VERSION[1]
  rescue => e
    return false
  end

  def unavailable?
    ret = !up_to_date? && !File.writable?(@db_dir)

    if ret && !@warned
      STDERR.puts "[You have no permission to update the pkgdb -- using a slow method]"
    end

    @warned = ret
  end

  def update_db(force = false)
    if !force
      up_to_date? and return false
    end

    close_db

    prev_sync = STDERR.sync
    STDERR.sync = true

    rebuild = force || !File.exist?(@db_file)

    STDERR.printf '[%s the pkgdb in %s ... ',
      rebuild ? 'Rebuilding' : 'Updating', @db_dir

    @installed_pkgs = installed_pkgs!.freeze

    if rebuild
      File.unlink(@db_file) if File.exist?(@db_file)

      @db = DBM.open(@db_filebase, 0664)

      new_pkgs = @installed_pkgs

      deleted_pkgs = []

      @installed_ports = []
    else
      @db = DBM.open(@db_filebase)

      db_mtime = date_db_file()

      s = @db[':installed_ports'] or return update_db(true)
      @installed_ports = s.split

      s = @db[':installed_pkgs'] or return update_db(true)
      prev_installed_pkgs = s.split

      new_pkgs = @installed_pkgs - prev_installed_pkgs
      deleted_pkgs = prev_installed_pkgs - @installed_pkgs

      (@installed_pkgs & prev_installed_pkgs).each do |pkg|
	pkg_mtime = date_installed(pkg)

	if db_mtime < pkg_mtime
	  new_pkgs << pkg
	  deleted_pkgs << pkg
	end
      end

      new_pkgs.sort!
      deleted_pkgs.sort!
    end

    STDERR.printf "- %d packages found (-%d +%d) ",
      @installed_pkgs.size, deleted_pkgs.size, new_pkgs.size

    unless deleted_pkgs.empty?
      STDERR.print '(...)'

      # NOTE: you cannot delete keys while you enumerate the database elements
      @db.select { |path, pkg|
	path[0] == ?/ && deleted_pkgs.qinclude?(pkg)
      }.each do |path, |
	@db.delete(path)
      end

      deleted_pkgs.each do |pkg|
	p_key = '?' + pkg
	origin = @db[p_key]
	@db.delete(p_key)

	if !origin
	  STDERR.print "(? #{pkg})"
	  next
	end

	o_key = '?' + origin

	pkgs = @db[o_key]

	if !pkgs
	  STDERR.print "(? #{origin})"
	  next
	end

	pkgs = pkgs.split
	pkgs.delete(pkg)

	if pkgs.empty?
	  @installed_ports.delete(origin)
	  @db.delete(o_key)
	else
	  @db[o_key] = pkgs.join(' ')
	end
      end
    end

    new_pkgs.each do |pkg|
      STDERR.putc ?.

      begin
	pkginfo = PkgInfo.new(pkg)

	if origin = pkginfo.origin
	  @installed_ports << origin

	  @db['?' + pkg] = origin

	  o_key = '?' + origin
	  o_val = @db[o_key]

	  if o_val
	    o_val = o_val.split << pkg
	    o_val.uniq!		# just in case
	    @db[o_key] = o_val.join(' ')
	  else
	    @db[o_key] = pkg
	  end
	end

	pkginfo.files.each do |path|
	  @db[path] = pkg
	end
      rescue => e
	STDERR.puts "", e.message + ": skipping..."
	next
      end
    end

    @installed_ports.sort!
    @installed_ports.uniq!

    @db[':db_version'] = Marshal.dump(DB_VERSION)
    @db[':installed_pkgs'] = @installed_pkgs.join(' ')
    @db[':installed_ports'] = @installed_ports.join(' ')
    @db[':mtime'] = Marshal.dump(Time.now)

    STDERR.puts " done]"

    close_db

    true
  rescue => e
    raise DBError, "#{e.message}: Cannot update the pkgdb!]"
  end

  def open_db
    @db and return @db

    update_db

    retried = false

    begin
      @db = DBM.open(@db_filebase)

      check_db_version or raise TypeError

      s = @db[':installed_pkgs']
      s.is_a?(String) or raise TypeError, 'installed_pkg - not a string'
      @installed_pkgs = s.split

      s = @db[':installed_ports']
      s.is_a?(String) or raise TypeError, 'installed_ports - not a string'
      @installed_ports = s.split
    rescue TypeError => e
      update_db(true)

      unless retried
	retried = true
	retry
      end

      raise DBError, "#{e.message}: Cannot read the pkgdb!"
    end

    @db
  end

  def close_db
    if @db
      @db.close
      @db = nil
    end
  end

  def which(path)
    path = File.expand_path(path)

    if unavailable?
      pkgname = `#{PkgDB::CMD[:pkg_info]} -qW #{path}`.strip

      if pkgname.empty?
	 nil
      else
	pkgname
      end
    else
      open_db

      @db[path]
    end
  rescue => e
    raise DBError, e.message
  end

  def date_installed(pkgname)
    installed?(pkgname) or return nil

    File.mtime(pkg_comment(pkgname)) ||
      File.mtime(pkg_descr(pkgname)) ||
      File.mtime(pkg_contents(pkgname)) rescue Time.at(0)
  end

  def installed_pkg?(pkgname)
    if unavailable?
      pkgdir?(pkgdir(pkgname))
    else
      installed_pkgs_().qinclude?(pkgname)
    end
  end

  alias installed? installed_pkg?

  def installed_port?(pkgname)
    installed_ports().qinclude?(pkgname)
  end

  def required?(pkgname)
    file = pkg_required_by(pkgname)

    File.exist?(file) && !File.zero?(file)
  end

  def required_by(pkgname)
    filename = pkg_required_by(pkgname)

    File.exist?(filename) or return nil

    deps = {}

    File.open(filename).each_line { |line|
      line.chomp!
      deps[line] = true unless line.empty?
    }

    deps.keys
  end

  def pkgdep(pkgname)
    filename = pkg_contents(pkgname)

    File.exist?(filename) or return nil

    deps = {}

    if File.size(filename) >= 65536	# 64KB
      open("| grep ^@pkgdep #{filename}")
    else
      File.open(filename)
    end.each { |line|
      if /^@pkgdep\s+(\S+)/ =~ line
	deps[$1] = true
      end
    }

    deps.keys
  end

  PKGDB_FILES.each_key do |key|
    module_eval %{
      def pkg_#{key.to_s}(pkgname)
	pkgfile(pkgname, #{key.inspect})
      end
    }
  end

  def self.parse_date(str)
    require 'parsedate'

    ary = ParseDate.parsedate(str)
    ary.pop
    tz = ary.pop

    case tz
    when 'GMT', 'UTC'
      Time.utc(*ary)
    else
      Time.local(*ary)
    end
  rescue
    raise ArgumentError, "#{str}: date format error"
  end

  def installed_pkgs!()
    Dir.entries(@db_dir).select { |pkgname| pkgdir?(pkgdir(pkgname)) }.sort
  rescue => e
    raise DBError, e.message
  end

  def installed_pkgs_()
    open_db if @installed_pkgs.nil?

    @installed_pkgs
  end

  def installed_pkgs()
    if unavailable?
      installed_pkgs!()
    else
      installed_pkgs_()
    end
  end

  def installed_ports!()
    ary = installed_pkgs!.map { |pkgname| origin(pkgname) }
    ary.sort!
    ary.uniq!
    ary
  end

  def installed_ports_()
    open_db if @installed_ports.nil?

    @installed_ports
  end

  def installed_ports()
    if unavailable?
      installed_ports!()
    else
      installed_ports_()      
    end
  end

  def glob(pattern = true, want_pkg_info = true)
    list = []
    pkg = nil

    case pattern
    when String
      # shortcut
      if pkg = pkg(pattern)
	if block_given?
	  yield(want_pkg_info ? pkg : pattern)
	  return nil
	else
	  return [want_pkg_info ? pkg : pattern]
	end
      end      

      if /^([<>]=?)(.*)/ =~ pattern
	op = $1
	arg = $2

	pkgs = glob(arg)

	begin
	  if pkgs.empty?
	    base = PkgDB.parse_date(arg)
	  else
	    pkgs.size >= 2 and
	      raise ArgumentError, "#{arg}: ambiguous package specification (#{pkgs.join(', ')})"

	    base = date_installed(pkgs[0].to_s)
	  end

	  pattern = op + base.strftime('%Y-%m-%d %H:%M:%S')
	rescue => e
	  STDERR.puts e.message

	  if block_given?
	    return nil
	  else
	    return []
	  end
	end
      end
    end

    installed_pkgs().each do |pkgname|
      pkg = PkgInfo.new(pkgname)

      pkg.match?(pattern) or next

      if block_given?
	yield(want_pkg_info ? pkg : pkgname)
      else
	list.push(want_pkg_info ? pkg : pkgname)
      end
    end

    if block_given?
      nil
    else
      list
    end
  end

  def each(want_pkg_info = true, &block)
    glob(true, want_pkg_info, &block)
  end

  def tsort(pkgnames)
    t = TSort.new

    pkgnames.each do |pkgname|
      deps = pkgdep(pkgname) || []

      t.add(pkgname, *deps)
    end

    t
  end

  def sort(pkgnames)
    tsort(pkgnames).tsort! & pkgnames
  end

  def sort!(pkgnames)
    pkgnames.replace(sort(pkgnames))
  end

  def recurse(pkgname, recurse_down = false, recurse_up = false, sanity_check = false)
    list = []

    if recurse_up || sanity_check and deps = pkgdep(pkgname)
      deps.each do |name|
	installed?(name) or
	  raise DBError, "Stale dependency: #{pkgname} --> #{name} -- manually run 'pkgdb -F' to fix, or specify -O to force."

	list << name if recurse_up
      end
    end

    list << pkgname

    if recurse_down || sanity_check and deps = required_by(pkgname)
      deps.each do |name|
	installed?(name) or
	  raise DBError, "Stale dependency: #{pkgname} <-- #{name} -- manually run 'pkgdb -F' to fix, or specify -O to force."

	list << name if recurse_down
      end
    end

    list
  end
end
