# $Idaemons: /home/cvs/pkgtools/lib/portsdb.rb,v 1.58 2001/11/16 18:21:06 knu Exp $

require 'singleton'
require 'tempfile'
require 'dbm' unless defined?(DBM)
require 'fnmatch' unless File.respond_to?(:fnmatch?)
require 'pkgmisc'

class PortsDB
  include Singleton
  include Enumerable

  DB_VERSION = [:FreeBSD, 2]

  PORTS_DIR = ENV['PORTSDIR'] || '/usr/ports'
  PORTS_DBDIR = ENV['PORTS_DBDIR'] || PORTS_DIR
  PORTS_DBDIR_LIST = [
    ENV['PORTS_DBDIR'],
    PORTS_DIR,
    PkgDB::DB_DIR,
    ENV['TMPDIR'],
    '/var/tmp',
    '/tmp'
  ].compact

  LANGUAGE_SPECIFIC_CATEGORIES = {
    "chinese"		=> "zh-",
    "french"		=> "fr-",
    "german"		=> "de-",
    "hebrew"		=> "iw-",
    "japanese"		=> "ja-",
    "korean"		=> "ko-",
    "russian"		=> "ru-",
    "ukrainian"		=> "uk-",
    "vietnamese"	=> "vi-",
  }

  MY_PORT = 'sysutils/portupgrade'

  attr_reader :ports_dir, :index_file, :db_dir, :db_file
  attr_accessor :ignore_categories, :extra_categories

  class IndexFileError < StandardError
#    def message
#      "index file error"
#    end
  end

  class DBError < StandardError
#    def message
#      "database file error"
#    end
  end

  def setup(db_dir = nil, ports_dir = nil)
    @ports_dir = ports_dir || PORTS_DIR
    @abs_ports_dir = nil
    @index_file = ENV['PORTS_INDEX'] || portdir('INDEX')
    @db = nil

    set_db_dir(db_dir || PORTS_DBDIR)

    @categories = nil
    @virtual_categories = nil
    @ignore_categories = []
    @extra_categories = []
    @origins = nil
    @pkgnames = nil
    @origins_by_categories = {}
    @ports = {}
    @localbase = nil
    @x11base = nil

    self
  end

  def set_db_dir(db_dir)
    @db_dir = db_dir

    @db_file = File.join(db_dir, 'INDEX.db')
    @db_filebase = @db_file.sub(/\.db$/, '')
    @rbo_file = File.join(db_dir, 'INDEX.rbo')

    close_db
  end
  private :set_db_dir

  def abs_ports_dir
    if @abs_ports_dir.nil?
      pwd = Dir.pwd
      Dir.chdir(@ports_dir)

      @abs_ports_dir = Dir.pwd

      Dir.chdir(pwd)
    end

    @abs_ports_dir
  end

  def localbase
    @localbase ||= `cd #{portdir(MY_PORT)} && make -V LOCALBASE`.strip
  end

  def x11base
    @x11base ||= `cd #{portdir(MY_PORT)} && make -V X11BASE`.strip
  end

  def join(category, port)
    File.join(category, port)
  end

  def split(origin)
    if %r"^([^./A-Z][^/]*)/([^./][^/]*)$" =~ path
      return $1, $2
    end

    nil
  end

  def strip(path, existing_only = false)
    path = path.tr_s('/', '/')

    %r"^(?:(/.+)/)?([^./][^/]*/[^./][^/]*)/?$" =~ path or return nil

    dir = $1
    port = $2

    if dir && dir != @ports_dir && dir != abs_ports_dir
      return nil
    end

    if existing_only && !exist?(port)
      return nil
    end

    port
  end

  def portdir(port)
    File.join(@ports_dir, port)
  end

  def subdirs(dir)
    %x"fgrep -v bsd.port.subdir.mk #{dir}/Makefile |
       make -f - -V SUBDIR 2> /dev/null".split.select { |i|
      File.directory?(File.join(dir, i))
    }.sort
  end

  def categories
    open_db if @categories.nil?

    @categories
  end

  def categories!
    customize_categories(subdirs(@ports_dir))
  end

  def customize_categories(cats)
    ((cats | @extra_categories) - @ignore_categories).sort
  end

  def category?(category)
    @categories.qinclude?(category)
  end

  def virtual_categories
    open_db if @virtual_categories.nil?

    @virtual_categories
  end

  def virtual_category?(category)
    @virtual_categories.qinclude?(category)
  end

  def update
    STDERR.print "Updating the ports index ..."
    STDERR.flush

    t = Tempfile.new('INDEX')
    t.close
    tmp = t.path

    dir = abs_ports_dir()

    if true
      system %`
	#{PkgDB::CMD[:make_describe_pass1]} |
	#{PkgDB::CMD[:make_describe_pass2]} |
	(cd #{dir} && perl Tools/make_index) |
	sed -e 's./.\001.g' |
	sort -t '|' +1 -2 |
	sed -e 's.\001./.g' > #{tmp}
      `
    else
      system %`
	cd #{dir} && (make describe ECHO_MSG='echo > /dev/null' 2> /dev/null |
	perl Tools/make_index) |
	sed -e 's./.\001.g' |
	sort -t '|' +1 -2 |
	sed -e 's.\001./.g' > #{tmp}
      `
    end

    if File.zero?(tmp)
      STDERR.puts 'failed to generate INDEX!'
      raise IndexFileError, "index generation error"
    end

    begin
      File.chmod(0644, tmp)
    rescue => e
      STDERR.puts e.message
      raise IndexFileError, "index chmod error"
    end

    if not system('/bin/mv', '-f', tmp, @index_file)
      STDERR.puts 'failed to overwrite INDEX!"'
      raise IndexFileError, "index overwrite error"
    end

    STDERR.puts " done"

    @categories = nil
    @virtual_categories = nil
    @origins = nil
    @pkgnames = nil
    @origins_by_categories = {}
    @ports = {}

    close_db
  end

  def open_db
    @db and return @db

    update_db

    retried = false

    begin
      File.open(@rbo_file) do |f|
	db_version = Marshal.load(f)

	if db_version[0] != DB_VERSION[0] || db_version[1] < DB_VERSION[1]
	  f.close
	  raise TypeError
	end

	@categories = customize_categories(Marshal.load(f))
	@virtual_categories = Marshal.load(f)
	@origins = Marshal.load(f)
	@pkgnames = Marshal.load(f)
	@origins_by_categories = Marshal.load(f)
	@ports = {}
      end
    rescue TypeError => e
      update_db(true)

      if retried
	raise DBError, "#{e.message}: Cannot read the ports DB! (#{@rbo_file})"
      else
	retried = true
	retry
      end
    end

    begin
      @db = DBM.open(@db_filebase)
    rescue => e
      raise DBError, "#{e.message}: Cannot open the ports DB! (#{@db_file})"
    end

    @db
  rescue => e
    STDERR.puts e.message
    raise DBError, 'database file error'
  end

  def close_db
    if @db
      @db.close
      @db = nil
    end
  end

  def date_index
    File.mtime(@index_file) rescue nil
  end

  def date_db
    d1 = File.mtime(@db_file) rescue nil
    d2 = File.mtime(@rbo_file) rescue nil

    if not d1 && d2
      d1 || d2
    elsif d1 < d2
      d1
    else
      d2
    end
  end

  def up_to_date?
    d1 = date_db() and d2 = date_index() and d1 >= d2
  end

  def update_db(force = false)
    update if not File.exist?(@index_file)

    if force
      close_db
    else
      up_to_date? and return false

      close_db

      File.writable?(@db_dir) or PORTS_DBDIR_LIST.each do |dir|
	set_db_dir(dir)

	up_to_date? and return false

	File.writable?(dir) and break
      end
    end

    prev_sync = STDERR.sync
    STDERR.sync = true

    STDERR.print "[Updating the ports database in #{@db_dir} ... "

    nports = `wc -l #{@index_file}`.to_i

    STDERR.printf "- %d port entries found ", nports

    i = -1

    @origins = []
    @pkgnames = []

    begin
      File.unlink(@db_file) if File.exist?(@db_file)

      @db = DBM.open(@db_filebase, 0664)

      File.open(@index_file) do |f|
	f.each_with_index do |line, i|
	  lineno = i + 1

	  if lineno % 100 == 0
	    if lineno % 1000 == 0
	      STDERR.print lineno
	    else
	      STDERR.putc(?.)
	    end
	  end

	  begin
	    port_info = PortInfo.new(line)
	    origin = port_info.origin
	    pkgname = port_info.pkgname

	    port_info.categories.each do |category|
	      if @origins_by_categories.key?(category)
		@origins_by_categories[category] << origin
	      else
		@origins_by_categories[category] = [origin]
	      end
	    end

	    @origins << origin
	    @pkgnames << pkgname

	    @db[origin] = port_info
	    @db[pkgname.to_s] = port_info
	  rescue => e
	    STDERR.puts @index_file + ":#{lineno}:#{e.message}"
	  end
	end
      end
    rescue => e
      File.unlink(@db_file) if File.exist?(@db_file)
      raise DBError, "#{e.message}: Cannot update the ports DB! (#{@db_file})]"
    ensure
      close_db
    end

    @categories = categories!
    @virtual_categories = (@origins_by_categories.keys - @categories).sort

    begin
      File.unlink(@rbo_file) if File.exist?(@rbo_file)

      STDERR.print ' (...)'

      File.open(@rbo_file, 'w') do |f|
	Marshal.dump(DB_VERSION, f)
	Marshal.dump(@categories, f)
	Marshal.dump(@virtual_categories, f)
	Marshal.dump(@origins, f)
	Marshal.dump(@pkgnames, f)
	Marshal.dump(@origins_by_categories, f)
      end
    rescue => e
      File.unlink(@rbo_file) if File.exist?(@rbo_file)
      raise DBError, "#{e.message}: Cannot update the ports DB! (#{@rbo_file})"
    end

    STDERR.puts " done]"
    STDOUT.sync = prev_sync

    true
  end

  def port(key)
    key.is_a?(PortInfo) and return key

    @ports[key] and return @ports[key]

    open_db

    line = @db[key]

    @ports[key] = if line then PortInfo.new(line) else nil end
  end

  alias [] port

  def ports(keys)
    keys.map { port(key) }
  end

  alias indices ports

  def origin(key)
    if p = port(key)
      p.origin
    else
      nil
    end
  end

  def origins(category = nil)
    open_db

    if category
      @origins_by_categories[category]
    else
      @origins
    end
  end

  def origins!(category = nil)
    if category
      # only lists the ports which primary category is the given category
      subdirs(portdir(category)).map { |i|
	File.join(category, i)
      }
    else
      list = []

      categories!.each do |i|
	list.concat(origins!(i))
      end

      list
    end
  end

  def each(category = nil)
    ports = origins(category) or return nil

    ports.each { |key|
      yield(@db[key])
    }
  end

  def each_category
    categories.each { |key|
      yield(key)
    }
  end

  def each_origin(category = nil)
    ports = origins(category) or return nil

    ports.each { |key|
      yield(key)
    }
  end

  def each_origin!(category = nil, &block)
    if category
      # only lists the ports which primary category is the given category
      subdirs(portdir(category)).each do |i|
	block.call(File.join(category, i))
      end
    else
      categories!.each do |i|
	each_origin!(i, &block)
      end
    end
  end

  def each_pkgname
    open_db

    @pkgnames.each { |key|
      yield(key)
    }
  end

  def glob(pattern = '*')
    list = []
    pkg = nil

    open_db

    case pattern
    when Regexp
      is_port = pattern.source.include?('/')
    else
      if /^[<>]/ =~ pattern
	raise "Invalid glob pattern: #{pattern}"
      end

      is_port = pattern.include?('/')

      # shortcut
      if portinfo = port(pattern)
	if block_given?
	  yield(portinfo)
	  return nil
	else
	  return [portinfo]
	end
      end
    end

    if is_port
      @origins.each do |origin|
	case pattern
	when Regexp
	  next if pattern !~ origin
	else
	  next if not File.fnmatch?(pattern, origin, File::FNM_PATHNAME)
	end

	if portinfo = port(origin)
	  if block_given?
	    yield(portinfo)
	  else
	    list.push(portinfo)
	  end
	end
      end
    else
      @pkgnames.each do |pkgname|
	next if not pkgname.match?(pattern)

	if portinfo = port(pkgname.to_s)
	  if block_given?
	    yield(portinfo)
	  else
	    list.push(portinfo)
	  end
	end
      end
    end

    if block_given?
      nil
    else
      list
    end
  rescue => e
    STDERR.puts e.message

    if block_given?
      return nil
    else
      return []
    end
  end

  def exist?(port, quick = false)
    return if %r"^[^/]+/[^/]+$" !~ port

    dir = portdir(port)

    return false if not File.file?(File.join(dir, 'Makefile'))

    return true if quick

    pkgname = `cd #{dir} && make -V PKGNAME 2>/dev/null`.chomp

    if pkgname.empty?
      false
    else
      pkgname
    end
  end

  def all_depends_list!(origin, before_args = nil, after_args = nil)
    `cd #{$portsdb.portdir(origin)} && #{before_args || ''} make #{after_args || ''} all-depends-list`.map { |line|
      strip(line.chomp, true)
    }.compact
  end

  def all_depends_list(origin, before_args = nil, after_args = nil)
    if !before_args && !after_args && i = port(origin)
      i.all_depends.map { |n| origin(n) }
    else
      all_depends_list!(origin, before_args, after_args)
    end
  end

  def masters(port)
    dir = portdir(port)

    ports = []

    `cd #{dir} ; make -dd -n 2>&1`.each do |line|
      if /^Searching for .*\.\.\.Caching .* for (\S+)/ =~ line.chomp
	path = File.expand_path($1)

	if (path.sub!(%r"^#{Regexp.quote(@ports_dir)}/", '') ||
	    path.sub!(%r"^#{Regexp.quote(abs_ports_dir)}/", '')) &&
	    exist?(path)
	  ports << path unless ports.include?(path)
	end
      end
    end

    ports.delete(port)

    ports
  end

  def latest_link(port)
    dir = portdir(port)

    flag, name = `cd #{dir} && make -V NO_LATEST_LINK -V LATEST_LINK`.to_a

    if flag.chomp.empty?
      name.chomp
    else
      nil
    end
  end

  def sort(ports)
    tsort = TSort.new

    ports.each do |p|
      portinfo = port(p)

      portinfo or next

      o = portinfo.origin

      deps = all_depends_list(o)	# XXX

      tsort.add(o, *deps)
    end

    tsort.tsort! & ports
  end

  def sort!(ports)
    ports.replace(sort(ports))
  end

  def recurse(portinfo, recurse_down = false, recurse_up = false)
    if not portinfo.is_a?(PortInfo)
      portinfo = port(portinfo)
    end

    list = []

    portinfo or return list

    if recurse_up
      portinfo.all_depends.map do |name|
	i = port(name)

	list << i if i
      end
    end

    list << portinfo

    if recurse_down
      # slow!
      pkgname = portinfo.pkgname.fullname

      glob do |i|
	list << i if i.all_depends.include?(pkgname)
      end
    end

    list
  end
end
